import asyncio
import json
import time
import traceback

from fastapi.encoders import jsonable_encoder
from httpx import ReadTimeout
from pydantic import BaseModel

from gravybox.betterstack import collect_logger
from gravybox.exceptions import GravyboxException, DataUnavailable
from gravybox.protocol import LinkRequest, Condition

logger = collect_logger()


def upstream_api_call(upstream_provider):
    """
    wrapper for all upstream api calls
    handles errors, task cancellations, metrics, and logging
    returns the result of the wrapped function, or a Condition if the wrapped function raises an exception
    """

    def decorator(function):
        async def wrapper(*args, link_request: LinkRequest = None, **kwargs):
            if link_request is None:
                raise ValueError("please pass the original link request when making a call to an upstream api")
            call_args = [arg for arg in args]
            call_kwargs = [f"{key}={value}" for key, value in kwargs.items()]
            log_extras = {
                "upstream_provider": upstream_provider,
                "upstream_call_type": function.__name__,
                "upstream_call_arguments": jsonable_encoder(call_args + call_kwargs),
                "trace_id": link_request.trace_id
            }
            logger.info("( ) calling upstream api", extra=log_extras)
            start_time = time.time()
            try:
                result: BaseModel = await function(*args, link_request=link_request, log_extras=log_extras, **kwargs)
                log_extras["result"] = result.model_dump_json()
                log_extras["condition"] = Condition.success.value
                return result
            except asyncio.CancelledError:
                log_extras["condition"] = Condition.cancelled.value
                raise
            except DataUnavailable:
                log_extras["condition"] = Condition.data_unavailable.value
                return Condition.data_unavailable
            except ReadTimeout:
                log_extras["condition"] = Condition.upstream_timeout.value
                return Condition.upstream_timeout
            except Exception as error:
                if isinstance(error, GravyboxException):
                    log_extras |= error.log_extras
                log_extras["error_str"] = str(error)
                log_extras["traceback"] = traceback.format_exc()
                log_extras["condition"] = Condition.unhandled_exception.value
                logger.error("(!) calling upstream api failed with unhandled exception", extra=log_extras)
                return Condition.unhandled_exception
            finally:
                log_extras["elapsed_time"] = time.time() - start_time
                logger.info("(*) completed upstream api call", extra=log_extras)

        return wrapper

    return decorator



from enum import Enum

from pydantic import BaseModel


class UpstreamCallParadigms(Enum):
    sequential = "sequential"
    simultaneous = "simultaneous"
    short_circuit = "short_circuit"


class Condition(Enum):
    success = "success"
    data_unavailable = "data unavailable"
    upstream_timeout = "upstream request timed out"
    unhandled_exception = "unhandled exception"
    invalid_request = "invalid request"
    cancelled = "cancelled"
    authentication_failure = "authentication failure"
    collection_failure = "collection failure"


class GravyboxRequest(BaseModel):
    pass


class GravyboxResponse(BaseModel):
    success: bool
    error: str = ""
    content: dict | None = None


class LinkRequest(GravyboxRequest):
    trace_id: str
    upstream_call_paradigm: UpstreamCallParadigms = UpstreamCallParadigms.short_circuit


class LinkResponse(GravyboxResponse):
    pass



from setuptools import setup

installation_requirements = [
    "logtail-python==0.2.10",
    "fastapi[standard]==0.112.2",
    "pydantic==2.8.2",
    "starlette==0.38.2",
    "httpx==0.27.2",
]

setup(
    name="gravybox",
    description="A big box of gravy for all of your FastAPI-ripping docker-aboded itty-bits. Enjoy at leisure.",
    version="0.85",
    url="https://github.com/clementinegroup/gravybox",
    author="(~)",
    package_dir={"": "packages"},
    packages=["gravybox"],
    install_requires=installation_requirements
)



import os

import httpx
from httpx import AsyncClient

from gravybox.betterstack import collect_logger

logger = collect_logger()

REQUEST_MANAGER_TIMEOUT = int(os.getenv("REQUEST_MANAGER_TIMEOUT", 60))


class AsyncRequestManager:
    """
    Singleton wrapper for an async http request client
    expects to be initialized once during FastAPI startup
    expects to be closed once during FastAPI shutdown
    use the static methods, avoid touching internal_* variables and functions
    can also be used as a context manager for test cases
    """

    def __init__(self):
        limits = httpx.Limits(max_keepalive_connections=None, max_connections=None, keepalive_expiry=None)
        timeout = httpx.Timeout(REQUEST_MANAGER_TIMEOUT)
        self.internal_request_client: AsyncClient = AsyncClient(timeout=timeout, limits=limits)

    async def __aenter__(self):
        logger.info("initializing request client")
        global INTERNAL_ASYNC_REQUEST_SINGLETON
        INTERNAL_ASYNC_REQUEST_SINGLETON = self
        return self

    async def __aexit__(self, exc_type, value, traceback):
        logger.info("shutting down request client")
        await self.client().aclose()

    @staticmethod
    def client() -> AsyncClient:
        global INTERNAL_ASYNC_REQUEST_SINGLETON
        if INTERNAL_ASYNC_REQUEST_SINGLETON is None:
            raise RuntimeError("AsyncClient.initialize() must be called before AsyncClient.client()")
        return INTERNAL_ASYNC_REQUEST_SINGLETON.internal_request_client

    @staticmethod
    def initialize():
        logger.info("initializing request client")
        global INTERNAL_ASYNC_REQUEST_SINGLETON
        INTERNAL_ASYNC_REQUEST_SINGLETON = AsyncRequestManager()

    @staticmethod
    async def shutdown():
        logger.info("shutting down request client")
        global INTERNAL_ASYNC_REQUEST_SINGLETON
        await INTERNAL_ASYNC_REQUEST_SINGLETON.internal_request_client.aclose()


INTERNAL_ASYNC_REQUEST_SINGLETON: AsyncRequestManager | None = None
